#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2022 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import json
import os.path

import cbor

from eventail.sync_publisher import Endpoint


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Publish a configuration event and its payload on the given broker"
    )
    parser.add_argument(
        "amqp_url", help="URL of the broker, including credentials", type=str
    )
    parser.add_argument("event", help="Configuration event name")
    parser.add_argument(
        "payload",
        help="The path to the file containing the payload, in JSON or CBOR format (from file extension).",
    )
    args = parser.parse_args()
    _, ext = os.path.splitext(args.payload)
    unserialize = json.loads if ext == ".json" else cbor.loads
    with open(args.payload, "rb") as ins:
        data = ins.read()
    payload = unserialize(data)

    endpoint = Endpoint(args.amqp_url, "debug_configuration_event_publisher")
    endpoint.publish_configuration(args.event, payload)



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018-2019 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import argparse
import time

import pika
from pika.exceptions import (
    ChannelClosed,
    ConnectionClosed,
    AMQPConnectionError,
    AMQPHeartbeatTimeout,
)


class Logger:

    LOG_EXCHANGE = "logs"
    LOG_EXCHANGE_TYPE = "topic"

    def __init__(self, url, routing_keys):
        connection = pika.BlockingConnection(pika.URLParameters(url))
        channel = connection.channel()

        channel.exchange_declare(
            exchange=self.LOG_EXCHANGE,
            exchange_type=self.LOG_EXCHANGE_TYPE,
            durable=True,
        )

        # We declare a transient queue because we don't want to fill-up rabbitmq
        # with logs if the logger is down
        result = channel.queue_declare("", exclusive=True)
        queue_name = result.method.queue

        for key in routing_keys:
            channel.queue_bind(exchange="logs", queue=queue_name, routing_key=key)
        # Logger queue is auto ack for minimum overhead as we don't care losing some
        # messages (very rare as we rarely fail)
        channel.basic_consume(
            queue=queue_name, on_message_callback=self.callback, auto_ack=True
        )
        self._channel = channel
        self._connection = connection

    def callback(self, ch, method, properties, body):
        print("[{}] {}".format(method.routing_key, body.decode("utf-8")))

    def run(self):
        try:
            self._channel.start_consuming()
        except KeyboardInterrupt:
            return True
        except (
            ChannelClosed,
            ConnectionClosed,
            AMQPConnectionError,
            AMQPHeartbeatTimeout,
        ):
            return False
        finally:
            if not self._connection.is_closed:
                self._connection.close()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Display selected logs in realtime on the given broker"
    )
    parser.add_argument("amqp_url", help="URL of the broker, including credentials")
    parser.add_argument(
        "--filter",
        help="Log patterns to subscribe to (default to all)",
        nargs="*",
        default=["#"],
    )
    args = parser.parse_args()
    expected_stop = False
    print("Ctrl-C to quit.")
    print("Subcribing to logs:", args.filter)
    while not expected_stop:
        try:
            logger = Logger(args.amqp_url, args.filter)
        except AMQPConnectionError:
            print("could not connect; retry…")
            time.sleep(2)
            continue
        print("connected!")
        expected_stop = logger.run()
    print("bye!")



#!/usr/bin/env python
#
# MIT License
#
# Copyright (c) 2018-2019 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
from typing import Any, Dict
import argparse
import os
import pprint

import cbor
from eventail.async_service.pika import Service


JSON_MODEL = Dict[str, Any]


class Monitor(Service):
    def __init__(self, save, *args, **kwargs):
        self.save = save
        super().__init__(*args, **kwargs)

    def dump(self, key, conversation_id, payload):
        if self.save:
            filename = f"{key}.{conversation_id}.cbor"
            if os.path.exists(filename):
                filename = f"{key}.{conversation_id}.2.cbor"
            with open(filename, "wb") as output:
                cbor.dump(payload, output)
            print("Payload saved as", filename)
        else:
            print("Payload:")
            pprint.pprint(payload)

    def handle_result(
        self,
        key: str,
        payload: JSON_MODEL,
        conversation_id: str,
        status: str,
        correlation_id: str,
        meta: Dict[str, str],
    ) -> None:
        print("Got a Result:", key)
        print("Conversation ID:", conversation_id)
        print("Correlation ID:", correlation_id)
        print("Status", status)
        print("Metadata:", meta)
        self.dump(key + "-result", conversation_id, payload)
        print("---------")
        print()

    def handle_command(
        self,
        command: str,
        payload: JSON_MODEL,
        conversation_id: str,
        reply_to: str,
        correlation_id: str,
        meta: Dict[str, str],
    ) -> None:
        print("Got a Command", command)
        print("Conversation ID:", conversation_id)
        print("Correlation ID", correlation_id)
        print("Return to", reply_to)
        print("Metadata:", meta)
        self.dump(command, conversation_id, payload)
        print("---------")
        print()

    def handle_event(
        self,
        event: str,
        payload: JSON_MODEL,
        conversation_id: str,
        meta: Dict[str, str],
    ) -> None:
        print("Got an Event", event)
        print("Conversation ID:", conversation_id)
        print("Metadata:", meta)
        self.dump(event, conversation_id, payload)
        print("---------")
        print()

    def handle_config(
        self,
        config: str,
        payload: JSON_MODEL,
        meta: Dict[str, str],
    ) -> bool:
        print("Got a configuration", config)
        print("Metadata:", meta)
        self.dump(config, "", payload)
        print("---------")
        print()
        return True

    def on_unconfigured(self):
        # Trick to unlock the monitor.
        self.publish_configuration("MonitorStarted", {})

    def on_ready(self) -> None:
        print("Started")


if __name__ == "__main__":
    # import logging
    # logger = logging.getLogger("async_service")
    # logger.addHandler(logging.StreamHandler())
    # logger.setLevel(logging.DEBUG)

    parser = argparse.ArgumentParser(
        description="Monitor selected Events and/or Commands on the given broker"
    )
    parser.add_argument("amqp_url", help="URL of the broker, including credentials")
    parser.add_argument(
        "--events",
        help="Event patterns to subscribe to (default to none)",
        nargs="*",
        default=["none"],
    )
    parser.add_argument(
        "--commands",
        help="Command patterns to subscribe to (default to none)",
        nargs="*",
        default=["none"],
    )
    parser.add_argument(
        "--configurations",
        help="Configuration patterns to subscribe to (default to none)",
        nargs="*",
        default=["none"],
    )
    parser.add_argument(
        "--save", action="store_true", help="save payloads in CBOR format."
    )
    args = parser.parse_args()
    monitor = Monitor(
        args.save,
        [args.amqp_url],
        args.events,
        args.commands,
        "debug_monitor",
        config_routing_keys=args.configurations + ["MonitorStarted"],
    )
    monitor.use_exclusive_queues()
    print("Subscribing to events:", args.events)
    print("Subscribing to commands:", args.commands)
    print("Subscribing to configurations:", args.configurations)
    print("Press Ctr-C to quit.")
    try:
        monitor.run()
    except KeyboardInterrupt:
        monitor.stop()



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="eventail",
    version="2.2.1",
    url="https://github.com/allo-media/eventail",
    author="Allo-Media",
    author_email="dev@allo-media.fr",
    description="A base class and utilities for AM service architecture",
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.8",
        "Intended Audience :: Developers",
        "Operating System :: POSIX :: Linux",
        "Topic :: Software Development",
    ],
    license="MIT",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    scripts=[
        "scripts/inspect_queue.py",
        "scripts/logger.py",
        "scripts/monitor.py",
        "scripts/publish_configuration.py",
        "scripts/publish_event.py",
        "scripts/resurrect.py",
        "scripts/send_command.py",
    ],
    install_requires=["pika>=1.2.0", "cbor", "redis"],
    extras_require={
        "asyncio": ["aiormq", "uvloop"],
        "synchronous": ["kombu"],
        "test": ["tox"],
    },
    python_requires=">=3.8",
)



#
# MIT License
#
# Copyright (c) 2018-2019 Groupe Allo-Media
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
"""
A base class implementing AM service architecture and its requirements for a synchronous publisher Endpoint.
"""
import logging
import os
import socket
from typing import Any, Dict, List, Optional

import cbor
from kombu import Connection, Exchange
from kombu.pools import producers, set_limit

from eventail.gelf import GELF

JSON_MODEL = Dict[str, Any]
LOGGER = logging.getLogger("sync_endpoint")
set_limit(2)  # two connections are more than enough; `async_service` uses just one.


class Endpoint:
    """A synchronous publishing endpoint for AlloMedia EDA."""

    ID = os.getpid()
    HOSTNAME = socket.gethostname()
    EVENT_EXCHANGE = "events"
    CONFIGURATION_EXCHANGE = "configurations"
    LOG_EXCHANGE = "logs"
    EVENT_EXCHANGE_TYPE = "topic"
    CONFIGURATION_EXCHANGE_TYPE = "topic"
    LOG_EXCHANGE_TYPE = "topic"

    def __init__(
        self,
        amqp_urls: List[str],
        logical_service: str,
        connection_max_retries: Optional[int] = None,
    ) -> None:
        """Initialize endpoint.

        * ``amqp_urls`` is a list of broker urls that will be tried in turn (round robin style) to send messages.
        * ``logical_service``: the logical service this endpoint belongs to.
        """

        self.logical_service = logical_service

        self._connection = Connection(
            amqp_urls,
            transport_options={
                "confirm_publish": True,
                "max_retries": connection_max_retries,
            },
        )
        self.event_exchange = Exchange(
            self.EVENT_EXCHANGE, type=self.EVENT_EXCHANGE_TYPE, durable=True
        )
        self.configuration_exchange = Exchange(
            self.CONFIGURATION_EXCHANGE,
            type=self.CONFIGURATION_EXCHANGE_TYPE,
            durable=True,
        )
        self.log_exchange = Exchange(
            self.LOG_EXCHANGE, self.LOG_EXCHANGE_TYPE, durable=True
        )
        self._force_json = False

    def force_json(self):
        """Force serialization of payload into JSON."""
        self._force_json = True

    def log(
        self,
        criticity: int,
        short: str,
        full: str = "",
        conversation_id: str = "",
        additional_fields: Dict = {},
    ) -> None:
        """Log to the log bus.

        Parameters:
         - `criticity`: int, in the syslog scale
         - `short`: str, short description of log
         - `full`: str, the full message of the log (appears as `message` in Graylog)
         - `additional_fields`: Dict, data to be merged into the GELF payload as additional fields
        """
        # no persistent messages, no retry
        gelf = GELF(self, criticity, short, full, conversation_id, additional_fields)
        LOGGER.debug("Application logged: %s\n%s", short, full)

        # the timeout here has no measurable effect in my tests, but I set it here to rule out
        # possible issues
        with producers[self._connection].acquire(block=True, timeout=2) as producer:
            producer.publish(
                gelf.payload,
                delivery_mode=1,  # not persistent
                exchange=self.log_exchange,
                serializer=None,
                routing_key=gelf.routing_key,
                declare=[self.log_exchange],
                retry=False,
            )

    def publish_configuration(
        self,
        event: str,
        message: JSON_MODEL,
        max_retries: Optional[int] = None,
    ) -> None:
        """Publish a configuration event on the bus.

        The ``event`` is the name of the event,
        and the `message` is any data conforming to the JSON model.
        `max_retries` is None by default (retry for ever) but can be an `int`
        to specify the number of attempts to send the message before the error is raised.
        The firt retry is immediate, then the interval increases by one second at each step.
        """

        with producers[self._connection].acquire(block=True, timeout=2) as producer:
            producer.publish(
                message if self._force_json else cbor.dumps(message),
                delivery_mode=1,  # not persistent
                exchange=self.configuration_exchange,
                routing_key=event,
                declare=[self.configuration_exchange],
                headers={},
                content_type=None if self._force_json else "application/cbor",
                content_encoding=None if self._force_json else "binary",
                serializer="json" if self._force_json else None,
                retry=True,
                retry_policy={
                    "interval_start": 0,  # First retry immediately,
                    "interval_step": 1,  # then increase by 1s for every retry.
                    "interval_max": 30,  # but don't exceed 30s between retries.
                    "max_retries": max_retries,
                },
            )

    def publish_event(
        self,
        event: str,
        message: JSON_MODEL,
        conversation_id: str,
        max_retries: Optional[int] = None,
    ) -> None:
        """Publish an event on the bus.

        The ``event`` is the name of the event,
        and the `message` is any data conforming to the JSON model.
        `max_retries` is None by default (retry for ever) but can be an `int`
        to specify the number of attempts to send the message before the error is raised.
        The firt retry is immediate, then the interval increases by one second at each step.
        """

        # the timeout here has no measurable effect in my tests, but I set it here to rule out
        # possible issues
        with producers[self._connection].acquire(block=True, timeout=2) as producer:
            headers = {"conversation_id": str(conversation_id)}
            producer.publish(
                message if self._force_json else cbor.dumps(message),
                delivery_mode=2,  # persistent
                exchange=self.event_exchange,
                routing_key=event,
                declare=[self.event_exchange],
                headers=headers,
                content_type=None if self._force_json else "application/cbor",
                content_encoding=None if self._force_json else "binary",
                serializer="json" if self._force_json else None,
                retry=True,
                retry_policy={
                    "interval_start": 0,  # First retry immediately,
                    "interval_step": 1,  # then increase by 1s for every retry.
                    "interval_max": 30,  # but don't exceed 30s between retries.
                    "max_retries": max_retries,
                },
            )
